use std::{fmt::Display, io::Write, time::Duration};

use async_tungstenite::{tokio::connect_async, tungstenite};
use color_eyre::{eyre::bail, Result};
use futures::{SinkExt, StreamExt};
use serde::{Deserialize, Serialize};
use serde_json::Value;
use thiserror::Error;
use tokio::time::timeout;
use tracing::{debug, error, trace};

use crate::output::{i3blocks::I3Blocks, waybar::Waybar, OutputFormat};

#[derive(Debug, Error)]
pub enum HomeAssistantError {
    #[error("empty message received")]
    EmptyMessage,
    #[error("unhandled message {0}")]
    UnhandledMessage(String),
    #[error("error deserializing message: {0}")]
    JsonError(#[from] serde_json::Error),
}

#[derive(Debug, Default)]
pub(crate) struct HomeAssistantBuilder {
    host: Option<String>,
    entity: Option<String>,
    token: Option<String>,
    insecure: bool,
    format: OutputFormat,
}

impl HomeAssistantBuilder {
    pub(crate) fn new() -> Self {
        Self::default()
    }

    pub(crate) fn host(mut self, host: String) -> Self {
        self.host = Some(host);
        self
    }

    pub(crate) fn entity(mut self, entity: String) -> Self {
        self.entity = Some(entity);
        self
    }

    pub(crate) fn token(mut self, token: String) -> Self {
        self.token = Some(token);
        self
    }

    pub(crate) fn insecure(mut self, insecure: bool) -> Self {
        self.insecure = insecure;
        self
    }

    pub(crate) fn format(mut self, format: OutputFormat) -> Self {
        self.format = format;
        self
    }

    pub(crate) fn build(self) -> HomeAssistant {
        HomeAssistant {
            host: self.host.unwrap(),
            entity: self.entity.unwrap(),
            insecure: self.insecure,
            token: self.token.unwrap(),
            format: self.format,
            id: 1,
            auth_complete: false,
        }
    }
}

pub(crate) struct HomeAssistant {
    token: String,
    entity: String,
    insecure: bool,
    host: String,
    id: u64,
    format: OutputFormat,
    auth_complete: bool,
}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct Message {
    #[serde(rename = "type")]
    message_type: MessageType,
    #[serde(skip_serializing_if = "Option::is_none")]
    ha_version: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    access_token: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    message: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    event_type: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    success: Option<bool>,
    #[serde(skip_serializing_if = "Option::is_none")]
    id: Option<u64>,
    #[serde(skip_serializing_if = "Option::is_none")]
    event: Option<Event>,
    #[serde(skip_serializing_if = "Option::is_none")]
    result: Option<Value>,
}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct Event {
    data: EventData,
    time_fired: String,
    origin: String,
    context: EventContext,
}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct EventData {
    new_state: EntityState,
    old_state: EntityState,
}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct EventContext {
    id: String,
    parent_id: Option<String>,
    user_id: Option<String>,
}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct EntityState {
    entity_id: String,
    last_changed: String,
    state: String,
    attributes: Value,
    last_updated: String,
    context: EventContext,
}

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
enum MessageType {
    AuthRequired,
    Auth,
    AuthOk,
    AuthInvalid,
    Result,
    SubscribeEvents,
    Event,
    SubscribeTrigger,
    UnsubscribeEvents,
    FireEvent,
    CallService,
    GetStates,
    GetConfig,
    GetServices,
    GetPanels,
    CameraThumbnail,
    MediaPlayerThumbnail,
    Ping,
    Pong,
    ValidateConfig,
}

impl Default for MessageType {
    fn default() -> Self {
        MessageType::Pong
    }
}

impl Display for MessageType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            MessageType::AuthRequired => write!(f, "AuthRequired"),
            MessageType::Auth => write!(f, "Auth"),
            MessageType::AuthOk => write!(f, "AuthOk"),
            MessageType::AuthInvalid => write!(f, "AuthInvalid"),
            MessageType::Result => write!(f, "Result"),
            MessageType::SubscribeEvents => write!(f, "SubscribeEvents"),
            MessageType::Event => write!(f, "Event"),
            MessageType::SubscribeTrigger => write!(f, "SubscribeTrigger"),
            MessageType::UnsubscribeEvents => write!(f, "UnsubscribeEvents"),
            MessageType::FireEvent => write!(f, "FireEvent"),
            MessageType::CallService => write!(f, "CallService"),
            MessageType::GetStates => write!(f, "GetStates"),
            MessageType::GetConfig => write!(f, "GetConfig"),
            MessageType::GetServices => write!(f, "GetServices"),
            MessageType::GetPanels => write!(f, "GetPanels"),
            MessageType::CameraThumbnail => write!(f, "CameraThumbnail"),
            MessageType::MediaPlayerThumbnail => write!(f, "MediaPlayerThumbnail"),
            MessageType::Ping => write!(f, "Ping"),
            MessageType::Pong => write!(f, "Pong"),
            MessageType::ValidateConfig => write!(f, "ValidateConfig"),
        }
    }
}

impl Message {
    fn auth() -> Self {
        Self {
            message_type: MessageType::Auth,
            ..Default::default()
        }
    }

    fn subscribe_events() -> Self {
        Self {
            message_type: MessageType::SubscribeEvents,
            ..Default::default()
        }
    }

    fn get_states() -> Self {
        Self {
            message_type: MessageType::GetStates,
            ..Default::default()
        }
    }

    fn to_json(&self) -> String {
        serde_json::to_string(&self).unwrap()
    }

    fn to_message(&self) -> tungstenite::Message {
        tungstenite::Message::from(self.to_json())
    }
}

impl HomeAssistant {
    pub(crate) fn builder() -> HomeAssistantBuilder {
        HomeAssistantBuilder::new()
    }

    fn incrementing_id(&mut self) -> u64 {
        let id = self.id;
        self.id = self.id.wrapping_add(1);
        id
    }

    pub(crate) async fn open_connection(mut self) -> Result<()> {
        let api_url = format!(
            "{}://{}/api/websocket",
            if self.insecure { "ws" } else { "wss" },
            self.host
        );

        debug!("Connecting to URL {api_url}");

        let (mut ws_stream, _) = connect_async(&api_url).await?;

        loop {
            match timeout(Duration::from_secs(5), ws_stream.next()).await {
                Ok(msg) => match msg {
                    Some(msg) => match msg {
                        Ok(msg) => match self.handle_message(msg).await {
                            Ok(response_messages) => {
                                for response in response_messages {
                                    ws_stream.send(response).await?
                                }
                            }
                            Err(e) => {
                                if let Some(err) = e.downcast_ref::<HomeAssistantError>() {
                                    match err {
                                        HomeAssistantError::EmptyMessage => {
                                            debug!("Received empty message, ignoring...")
                                        }
                                        HomeAssistantError::UnhandledMessage(message_type) => {
                                            debug!(
                                                    "Received '{message_type}', we're currently not handling that."
                                                )
                                        }
                                        HomeAssistantError::JsonError(e) => {
                                            error!("Error deserializing JSON: {e}")
                                        }
                                    }
                                } else {
                                    error!("{e}");
                                }
                            }
                        },
                        Err(e) => {
                            error!("Error in receiving message: {e}");
                            // break;
                        }
                    },
                    None => {
                        // I recon we should reconnect here. Probably.
                        error!("Received 'None' from handle_message");
                        break;
                    }
                },
                Err(e) => {
                    error!("Timeout reading message: {e}");
                    // return Err(HomeAssistantError::EmptyMessage);
                    bail!(e);
                }
            };
        }

        Ok(())
    }

    async fn handle_message(
        &mut self,
        msg: tungstenite::Message,
    ) -> Result<Vec<tungstenite::Message>> {
        if msg.is_empty() {
            bail!(HomeAssistantError::EmptyMessage);
        }

        trace!("{}", msg.clone().into_text()?);
        let message: Message = serde_json::from_slice(&msg.into_data())?;
        trace!("Received: {message:?}");

        let response = match message.message_type {
            MessageType::AuthRequired => {
                debug!("Sending Authentication Data");
                let mut auth = Message::auth();
                auth.access_token = Some(self.token.clone());

                vec![auth.to_message()]
            }
            MessageType::AuthOk => {
                let mut subscribe = Message::subscribe_events();
                subscribe.event_type = Some("state_changed".into());
                subscribe.id = Some(self.incrementing_id());

                let mut get_states = Message::get_states();
                get_states.id = Some(self.incrementing_id());

                self.auth_complete = true;

                vec![subscribe.to_message(), get_states.to_message()]
            }
            MessageType::Event => {
                let event = message.event.unwrap();

                if event.data.new_state.entity_id == self.entity {
                    let output = self
                        .format_state(event.data.new_state.state, event.data.new_state.attributes);
                    println!("{output}");
                    std::io::stdout().flush()?;
                }

                vec![]
            }
            MessageType::Result => {
                trace!("Result: {:?}", message);

                if let Some(result) = message.result {
                    if let Some(items) = result.as_array() {
                        if let Some(entity) = items.iter().find(|&item| {
                            item.as_object()
                                .unwrap()
                                .get::<String>(&String::from("entity_id"))
                                .unwrap()
                                .eq(&self.entity)
                        }) {
                            let state: EntityState =
                                serde_json::from_value(entity.clone()).unwrap();
                            println!("{}", self.format_state(state.state, state.attributes));
                            std::io::stdout().flush()?;
                        }
                    }
                }

                vec![]
            }
            _ => bail!(HomeAssistantError::UnhandledMessage(
                message.message_type.to_string()
            )),
        };
        Ok(response)
    }

    fn format_state(&self, state: String, attributes: Value) -> String {
        if state == "playing" {
            let maybe_channel = attributes["media_channel"].as_str();
            let artist = attributes["media_artist"].as_str().unwrap_or("No artist");
            let title = attributes["media_title"].as_str().unwrap_or("No title");
            let volume_raw = attributes["volume_level"].as_f64().unwrap_or(-1.);
            let is_muted = attributes["is_volume_muted"].as_bool().unwrap_or(false);

            let volume = if is_muted {
                String::from("[  \u{f466} ] ")
            } else if volume_raw >= 0. {
                format!("[{:3.0}%] ", volume_raw * 100.)
            } else {
                String::new()
            };

            let text = if let Some(channel) = maybe_channel {
                format!("{volume}[{channel}] {artist} - {title}")
            } else {
                format!("{volume}{artist} - {title}")
            };

            // let text = html_escape::encode_text(&now_playing).to_string();

            match self.format {
                OutputFormat::Waybar => Waybar::builder()
                    .text(text)
                    .percentage((volume_raw * 100.) as u8)
                    .muted(is_muted)
                    .build()
                    .to_string(),
                OutputFormat::I3Blocks => I3Blocks::builder().text(text).build().to_string(),
            }
        } else {
            // println!("Sonos {state}");
            match self.format {
                OutputFormat::Waybar => Waybar::builder()
                    .text(format!("Sonos {state}"))
                    .build()
                    .to_string(),
                OutputFormat::I3Blocks => I3Blocks::builder()
                    .text(format!("Sonos {state}"))
                    .build()
                    .to_string(),
            }
        }
    }
}
